import logging
import multiprocessing

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy as sp
import seaborn as sns

from sklearn import (
    cross_validation, ensemble, grid_search, learning_curve, linear_model,
    metrics, naive_bayes, pipeline, preprocessing)

from imblearn.metrics import specificity_score


###############
# Classifiers #
###############

CLASSIFIERS = {
    'ExtraTrees': {
        'name': 'Extra Randomized Trees',
        'class': ensemble.ExtraTreesClassifier,
        'best_params': {  # MCC: 0.582
            'max_depth': 9,
            'criterion': 'entropy',
            'min_samples_leaf': 7,
            'bootstrap': True,
            'max_features': 0.47,
            'min_samples_split': 4,
            'n_jobs': 2
        },
        'param_distributions': {
            'max_depth': sp.stats.randint(5, 10),
            'max_features': sp.stats.uniform(0.05, 0.5),
            'min_samples_split': sp.stats.randint(1, 8),
            'min_samples_leaf': sp.stats.randint(6, 15),
            'bootstrap': [True, False],
            'criterion': ['gini', 'entropy'],
            'n_jobs': [2]
        }
    },
    'RandomForest': {
        'name': 'Random Forest',
        'best_params': {  # MCC: 0.638
            'bootstrap': True,
            'criterion': 'gini',
            'max_depth': 9,
            'max_features': 0.49,
            'min_samples_leaf': 10,
            'min_samples_split': 2,
            'n_estimators': 13,
            'n_jobs': 1,
        },
        'class': ensemble.RandomForestClassifier,
        'param_distributions': {
            'n_estimators': sp.stats.randint(5, 20),
            'criterion': ['gini', 'entropy'],
            'max_features': sp.stats.uniform(0.05, 0.5),
            'min_samples_split': sp.stats.randint(1, 8),
            'min_samples_leaf': sp.stats.randint(6, 15),
            'bootstrap': [True, False],
            'n_jobs': [2]
        }
    },
    'AdaBoost': {
        'name': 'Ada Boost',
        'class': ensemble.AdaBoostClassifier,
        'best_params': {  # MCC: 0.641
            'n_estimators': 258
        },
        'param_distributions': {
            'n_estimators': sp.stats.randint(50, 300)
        }
    },
    'LogisticRegression': {
        'name': 'Logistic Regression',
        'class': linear_model.LogisticRegression,
        'best_params': {  # MCC: 0.487
            'C': 1.425494495402806,
            'dual': False,
            'fit_intercept': True,
            'max_iter': 146,
            'n_jobs': 2,
            'penalty': 'l2',
            'solver': 'lbfgs'
        },
        'param_distributions': {
            'penalty': ['l2'],  # (default)
            'dual': [False],  # (default), better when # samples > # features
            'C': sp.stats.uniform(0.5, 1.5),  # (default: 1.0)
            'fit_intercept': [True],  # (default), add biast to decision
            'solver': ['newton-cg', 'lbfgs', 'liblinear', 'sag'],  # liblinear
            'max_iter': sp.stats.randint(100, 250),  # (default: 100)
            'n_jobs': [2]
        }
    },
    'SGD': {
        'name': 'Stochastic Gradient Descent',
        'class': linear_model.SGDClassifier,
        'best_params': {  # MCC: 0.488
            'loss': 'log',
            'n_iter': 11,
            'penalty': 'l1'
        },
        'param_distributions': {
            'loss': ['hinge', 'log', 'modified_huber', 'perceptron', 'huber'],
            'penalty': ['l2', 'l1', 'elasticnet'],
            'n_iter': sp.stats.randint(6, 15)
        }
    },
    'NaiveBayes': {
        'name': 'Gaussian Naive Bayes',
        'class': naive_bayes.GaussianNB,
        'best_params': {},  # No parameters
        'param_distributions': {}
    }
}

# Set classifier colors
for classifier_name, color in zip(sorted(CLASSIFIERS), sns.color_palette()):
    CLASSIFIERS[classifier_name]['color'] = color

# Swap nicer colors
_red = CLASSIFIERS['LogisticRegression']['color']
_gold = CLASSIFIERS['RandomForest']['color']
CLASSIFIERS['LogisticRegression']['color'] = _gold
CLASSIFIERS['RandomForest']['color'] = _red


def _get_classifier(classifier_name, classifier_params=None):
    if classifier_params:
        return CLASSIFIERS[classifier_name]['class'](**classifier_params)
    else:
        return CLASSIFIERS[classifier_name]['class'](
            **CLASSIFIERS[classifier_name]['best_params'])


def _get_estimator(classifier_name, classifier_params=None):
    classifier = _get_classifier(classifier_name, classifier_params)
    return pipeline.Pipeline([
        ("imputer", preprocessing.Imputer(strategy="median")),
        ("scaler", preprocessing.StandardScaler()),
        ("classifier", classifier)])


#########
# Folds #
#########

FOLDS = {
    'kfold': {
        'name': 'k-fold',
        'class': cross_validation.KFold,
        'params': {
            'n_folds': 10,
            'shuffle': True,
        }
    },
    'stratified_kfold': {
        'name': 'Stratified k-fold',
        'class': cross_validation.StratifiedKFold,
        'params': {
            'n_folds': 10,
            'shuffle': True,
        }
    },
    'label_kfold': {
        'name': 'Label k-fold',
        'class': cross_validation.LabelKFold,
        'params': {
            'n_folds': 10,
        }
    },
    'predefined': {
        'name': 'Predefined split',
        'class': cross_validation.PredefinedSplit,
        'params': {}
    }
}


def _get_folds(folds_name, y, folds_params=None, labels=None):
    # Get default params
    params = FOLDS[folds_name]['params']

    # Update params if given
    if folds_params:
        params.update(folds_params)

    # Set default params depending on y and labels
    if folds_name == 'kfold':
        params['n'] = len(y)
    elif folds_name == 'stratified_kfold':
        params['y'] = y
    elif folds_name == 'label_kfold':
        params['labels'] = labels

    # Get folds
    return FOLDS[folds_name]['class'](**params)


def _evaluate(X, y, estimator, train, test, percentiles=None, scores=None):
    if not percentiles:
        percentiles = [100]

    estimator.fit(X[train], y[train])
    prob = estimator.predict_proba(X[test])[:, 1]
    pred = prob > 0.5

    results = pd.DataFrame()

    for percentile in percentiles:
        # Get confidence percentiles
        #low_score = np.percentile(prob, percentile/2.0)
        #high_score = np.percentile(prob, 100 - percentile/2.0)
        low_score = np.percentile(prob[prob < 0.5], percentile)
        high_score = np.percentile(prob[prob > 0.5], 100 - percentile)

        # Filter predictions with higher confidence
        fil = np.logical_or(prob <= low_score, high_score <= prob)

        # Initialize evaluation series
        evaluation = pd.Series((
            percentile,
            len(X),
            len(train),
            100.*len(train)/len(X),
            len(test),
            100.*sum(fil)/len(X),
            low_score,
            high_score,
            y,
            train,
            test,
            prob,
            pred,
        ), index=(
            'percentile', 'n_variants', 'n_train', 'p_train', 'n_test',
            'p_test', 'low_score', 'high_score',
            'y', 'train', 'test', 'prob', 'pred')
        )

        # Append evaluation
        evaluation = evaluation.append(
            evaluate_prediction(y[test[fil]], pred[fil]))

        # Append ROC curve
        if not (y[test[fil]].all() or (~y[test[fil]]).all()):
            evaluation = evaluation.append(
                pd.Series(metrics.roc_curve(y[test[fil]], prob[fil]),
                          index=('fpr', 'tpr', 'thresholds')))

        results = results.append(evaluation, ignore_index=True)

    return results


##################
# Cross-validate #
##################

def cross_validate(
        variants,
        classifiers=[('RandomForest', None)],
        folds=[('kfold', None)],
        label_name=None,
        scores=None,
        percentiles=None):
    # Get X, y, labels and estimators
    X, y = _get_X_y(variants)
    labels = variants[label_name] if label_name else None
    estimators = [
        _get_estimator(classifier_name, classifier_params)
        for classifier_name, classifier_params in classifiers]

    results = []
    for folds_name, folds_params in folds:
        logging.info('Generating %s (%s)...', folds_name, str(folds_params))
        for i, (train, test) in enumerate(
                _get_folds(folds_name, y, folds_params, labels)):
            print('.', end=' ')
            for (classifier_name, _), estimator in \
                    zip(classifiers, estimators):
                result = _evaluate(
                    X, y, estimator, train, test, percentiles, scores)
                result['classifier'] = CLASSIFIERS[classifier_name]['name']
                result['folds'] = FOLDS[folds_name]['name'] if folds_name != 'label_kfold' else label_name + ' k-fold'
                results.append(result)

    return pd.concat(results)


def roc_curve(cv_results):
    sns.set_style("whitegrid")
    sns.set_context("poster")
    figure = plt.figure()

    for cv_i, ((classifier, folds, percentile), cv_group) in enumerate(cv_results.groupby(
            ['classifier', 'folds', 'percentile'])):

        mean_tpr = 0.0
        mean_fpr = np.linspace(0, 1, 100)

        for i, result in cv_group.iterrows():
            mean_tpr += sp.interp(mean_fpr, result.fpr, result.tpr)

        mean_tpr /= len(cv_group)
        mean_tpr[0], mean_tpr[-1] = 0.0, 1.0

        color = next(v['color'] for k, v in CLASSIFIERS.items()
                     if v['name'] == classifier)
        # QUICKFIX: get always different colors
        color = sns.color_palette()[cv_i % len(sns.color_palette())]

        if percentile == 100:
            line_style = '-'
        elif percentile == 66:
            line_style = '--'
        else:
            line_style = ':'

        plt.plot(
            mean_fpr,
            mean_tpr,
            lw=1.8,
            color=color,
            ls=line_style,
            label='%s. %d%%, %s. AUC = %0.3f, MCC = %0.3f' % (
                classifier,
                percentile,
                folds,
                cv_group.auc.mean(),
                cv_group.mcc.mean()))

    plt.plot([0, 1], [0, 1], '--', color=(0.6, 0.6, 0.6, 0.7), label='Luck')
    plt.xlim([0, 1])
    plt.ylim([0, 1])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver operating characteristic')
    plt.legend(loc='lower right', prop={'size': 16})

    return figure


######################
# Features selection #
######################

def new_feature_score(params):
    indices, folds, clf, X, y = params
    return cross_validation.cross_val_score(
        clf, X[:, indices], y, cv=folds,
        scoring=metrics.make_scorer(metrics.matthews_corrcoef)
    ).mean()


def add_feature(indices, folds, estimator, X, y, n_jobs=4):
    new_indices = list(set(range(X[0].size)) - set(indices))
    params_list = [
        (indices + [i], folds, estimator, X, y)
        for i in new_indices
    ]

    pool = multiprocessing.Pool(n_jobs)
    scores = pool.map(new_feature_score, params_list)

    indices.append(new_indices[scores.index(max(scores))])
    return max(scores)


def remove_feature(indices, folds, estimator, X, y):
    estimator.fit(X[:, indices], y)
    importances = list(
        estimator.named_steps['classifier'].feature_importances_)
    del indices[importances.index(min(importances))]
    return indices


def iterative_features_selection(
        variants, max_features, folds_name='kfold', label_name=None,
        features=[], n_jobs=1):
    X, y = _get_X_y(variants)
    estimator = _get_estimator('RandomForest')
    labels = variants[label_name] if label_name else None

    # Best feature indices
    all_features = np.array(variants.filter(like='feature_').columns)
    indices = [list(all_features).index(feature) for feature in features]

    mcc = metrics.make_scorer(metrics.matthews_corrcoef)

    for i in range(max_features):
        # Folds
        folds = _get_folds(folds_name, y, labels=labels)

        # All features
        score = cross_validation.cross_val_score(
            estimator, X, y, scoring=mcc, cv=folds, n_jobs=n_jobs).mean()
        print('MCC = %.3f. All features (%d)' % (score, len(all_features)))

        # Currently selected features
        if indices:
            score = cross_validation.cross_val_score(
                estimator, X[:, indices], y, scoring=mcc, cv=folds,
                n_jobs=n_jobs).mean()
            print('MCC = %.3f. Current features (%d: %s)' % (
                score, len(indices), str(all_features[indices])))

        # Add three features
        score = add_feature(indices, folds, estimator, X, y, n_jobs)
        print('MCC = %.3f. Added 1 feature (%d: %s)' % (
            score, len(indices), str(all_features[indices])))
        score = add_feature(indices, folds, estimator, X, y, n_jobs)
        print('MCC = %.3f. Added 2 features (%d: %s)' % (
            score, len(indices), str(all_features[indices])))
        score = add_feature(indices, folds, estimator, X, y, n_jobs)
        print('MCC = %.3f. Added 3 features (%d: %s)' % (
            score, len(indices), str(all_features[indices])))

        # Remove one feature
        indices = remove_feature(indices, folds, estimator, X, y)
        score = cross_validation.cross_val_score(
            estimator, X[:, indices], y, scoring=mcc, cv=folds,
            n_jobs=n_jobs).mean()
        print('MCC = %.3f. Removed one feature (%d: %s)' % (
            score, len(indices), str(all_features[indices])))


##################
# Learning curve #
##################

def get_learning_curve(
        variants,
        classifier_name='RandomForest', classifier_params=None,
        folds_name='kfold', folds_params=None, label_name=None):
    # Get X, y, labels, estimator, folds and scorer
    X, y = _get_X_y(variants)
    labels = variants[label_name] if label_name else None
    estimator = _get_estimator(classifier_name, classifier_params)
    cv = _get_folds(folds_name, y, folds_params, labels)
    mcc_scorer = metrics.make_scorer(metrics.matthews_corrcoef)

    train_sizes, train_scores, test_scores = learning_curve.learning_curve(
        estimator, X, y, cv=cv, scoring=mcc_scorer)

    figure = plt.figure()
    plt.title('Learning curve')
    plt.xlabel('Training examples')
    plt.ylabel('MCC score')
    train_scores_mean = np.mean(train_scores, axis=1)
    train_scores_std = np.std(train_scores, axis=1)
    test_scores_mean = np.mean(test_scores, axis=1)
    test_scores_std = np.std(test_scores, axis=1)
    plt.grid()
    plt.fill_between(train_sizes, train_scores_mean - train_scores_std,
                     train_scores_mean + train_scores_std, alpha=0.1,
                     color="r")
    plt.fill_between(train_sizes, test_scores_mean - test_scores_std,
                     test_scores_mean + test_scores_std, alpha=0.1, color='g')
    plt.plot(train_sizes, train_scores_mean,
             'o-', color='r', label='Training score')
    plt.plot(train_sizes, test_scores_mean,
             'o-', color='g', label='Cross-validation score')
    plt.legend(loc='best')

    return figure


############
# Evaluate #
############

def evaluate(variants, estimator):
    """
    Evaluate the predictions of an estimator over an annotated set of variants

    variants must have a disease column
    """
    predict(variants, estimator)
    return evaluate_prediction(
        variants.disease, variants.pred_disease)


def evaluate_prediction(y, pred):
    # Convert to numpy array
    pred = np.array(pred, dtype=np.float64)

    # Count null predictions
    n_null = np.isnan(pred).sum()
    #pred.isnull().sum()

    # Remove null predictions and convert to np.bool
    y = y[~np.isnan(pred)].astype(np.bool)
    pred = pred[~np.isnan(pred)].astype(np.bool)

    if y.all() or (~y).all():
        return pd.Series((
            1.0 - n_null/(n_null + len(y)),
            metrics.accuracy_score(y, pred),
            np.nan, np.nan, np.nan, np.nan,
        ), index=('coverage', 'accuracy', 'specificity', 'sensitivity', 'auc', 'mcc'))
    return pd.Series((
        1.0 - n_null/(n_null + len(y)),
        metrics.accuracy_score(y, pred),
        specificity_score(y, pred),
        metrics.recall_score(y, pred),
        metrics.roc_auc_score(y, pred),
        metrics.matthews_corrcoef(y, pred),
    ), index=('coverage', 'accuracy', 'specificity', 'sensitivity', 'auc', 'mcc'))


################################
# Classifiers parameter search #
################################

def random_parameter_search(
        variants,
        n_iter=10,
        classifier_name='RandomForest',
        folds_name='kfold',
        folds_params=None,
        label_name=None):
    # Get X, y and classifier
    X, y = _get_X_y(variants)
    estimator = _get_estimator(classifier_name)

    # Get labels if given
    labels = variants[label_name] if label_name else None

    # Get folds
    logging.info('Generating %s folds, params: %s',
                 folds_name, str(folds_params))
    folds = _get_folds(folds_name, y, folds_params, labels)

    # User Mathews Correlation Coefficient scorer
    mcc_scorer = metrics.make_scorer(metrics.matthews_corrcoef)

    # Get parameters distribution for estimator
    param_distributions = {
        'classifier__' + k: v
        for k, v in CLASSIFIERS[classifier_name]['param_distributions'].items()
    }

    search = grid_search.RandomizedSearchCV(
        estimator,
        param_distributions,
        cv=folds,
        n_iter=n_iter,
        scoring=mcc_scorer,
        verbose=1)

    search.fit(X, y)
    return search


#########
# Train #
#########

def train(variants, classifier_name='RandomForest', classifier_params=None):
    """
    Train an estimator given the variants training set,
    the classifier name and its parameters
    """
    # Get X, y, estimator, fit and return
    X, y = _get_X_y(variants)
    estimator = _get_estimator(classifier_name, classifier_params)
    estimator.fit(X, y)
    return estimator


###########
# Predict #
###########

def predict(variants, estimator):
    """Get the prediction for variants using the given estimator"""
    X = _get_X(variants)
    variants['pred_score'] = estimator.predict_proba(X)[:, 1]
    variants['pred_disease'] = variants['pred_score'] > 0.5


def _get_X(variants):
    """Get features matrix X from variants"""
    X = variants.filter(like='feature_').as_matrix()
    X[X == -np.inf] = np.nan
    X[X == np.inf] = np.nan
    return X


def _get_X_y(variants):
    """Get features matrix X and disease vector y from variants"""
    X = _get_X(variants)
    y = variants.disease.as_matrix().astype(bool)
    return X, y



# -*- coding: utf-8 -*-

import collections
import gzip
import itertools
import io
import logging
import math
import re
import sys

import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt
plt.switch_backend('agg')

import numpy as np
import seaborn as sns

from .amino_acids import _SUBSTITUTION_MATRICES
from .data import path_blast, path_kalign, uniref100, get_network_parameters
from .utils import load_fasta_iter

############
# Features #
############

_DBS = ['100', '90']

_FEATURES = {
    'basic': ['hkd', 'pos'] +
             ['blosum50', 'blosum62', 'blosum80', 'miyata', 'pam60'] +
             [''.join(l) for l in itertools.product(
                 ['hwo', 'hww', 'vol'], ['_d', '_r'])],
    'topology': ['top_' + s for s in [
                 'betweenness', 'cc', 'closeness', 'degree',
                 'degree_centrality', 'eigenvector']],
    'blast': [''.join(l) for l in itertools.product(
              ['b'],
              [d[0] for d in _DBS],
              ['_all', '_eva', '_hum', '_nhu'],
              ['_nwt', '_nmt', '_naa', '_nal', '_rwt', '_rmt', '_pwm'],
              ['', '_w'])],
    'kalign': [''.join(l) for l in itertools.product(
               ['k'],
               [d[0] for d in _DBS],
               ['_all', '_hum', '_nhu'],
               ['_nwt', '_nmt', '_naa', '_nal', '_rwt', '_rmt', '_pwm',
                '_dh', '_dp', '_de', '_dv', '_da', '_db'],
               ['', '_w'])]
}

# Prepend 'feature_'
_FEATURES = {k: ['feature_' + x for x in v] for k, v in _FEATURES.items()}

FEATURES = tuple(itertools.chain(*_FEATURES.values()))

_BLAST_FEATURES = {
    d: set([f for f in _FEATURES['blast'] if f.startswith('feature_b' + d[0])])
    for d in _DBS
}

_KALIGN_FEATURES = {
    d: set([f for f in _FEATURES['kalign']
            if f.startswith('feature_k' + d[0])])
    for d in _DBS
}

PMUT_FEATURES = sorted([
    'feature_k1_all_nwt_w',
    'feature_b1_eva_rwt',
    'feature_k1_all_nal',
    'feature_k9_all_nwt_w',
    'feature_b1_eva_naa',
    'feature_k9_hum_naa_w',
    'feature_b9_eva_naa',
    'feature_k9_all_naa_w',
    'feature_k1_all_pwm_w',
    'feature_b1_eva_pwm_w',
    'feature_k1_hum_naa',
    'feature_miyata',
])

#################
# Feature names #
#################

# TODO: update feature names

_FEATURE_NAMES = {
    'feature_blosum62': 'BLOSUM 62',
    'feature_hwo_d': 'Hidrophobicity difference',
    'feature_top_degree': 'Topology Degree',
    'feature_b1_eva_nwt': 'BLAST UniRef100 # wt',
    'feature_b1_eva_nmt_w': 'BLAST UniRef100 % mt',
    'feature_b1_eva_naa': 'BLAST UniRef100 # aas',
    'feature_b1_eva_pwm': 'BLAST UniRef100 PSSM',
    'feature_b1_eva_pwm_w': 'BLAST UniRef100 weighted PSSM',
    'feature_k1_all_nwt': 'Kalign UniRef100 # wt',
    'feature_k1_all_nal': 'Kalign UniRef100 # align',
    'feature_k1_all_pwm_w': 'Kalign UniRef100 weighted PSSM',
    'feature_k1_hum_nmt_w': 'Kalign Human UniRef100 weighted # mt',
    'feature_k1_hum_nal': 'Kalign Human UniRef100 # align',
    'feature_k9_all_nwt_w': 'Kalign UniRef90 weighted # wt',
    'feature_k9_hum_naa': 'Kalign Human UniRef90',
}


def feature_name(feature):
    """Return a human readable name of the feature"""
    if feature in _FEATURE_NAMES:
        return _FEATURE_NAMES[feature]
    return feature


########################
# Features computation #
########################

def compute_features(variants, features=None):
    """Compute features of variants"""

    if features is None:
        features = FEATURES

    _basic_features(variants, features)
    _topology_features(variants, features)
    _blast_features(variants, features)
    _kalign_features(variants, features)


##################
# Basic features #
##################

# Average volume of buried residues
# Source: Amino Acid Volumes. -C. Chothia, Nature 254:304-308(1975)
_AA_Vol = {
    'A':  92., 'C': 106., 'D': 125., 'E': 155., 'F': 203., 'G':  66.,
    'H': 167., 'I': 169., 'K': 171., 'L': 168., 'M': 171., 'N': 125.,
    'P': 129., 'Q': 161., 'R': 202., 'S':  99., 'T': 122., 'V': 142.,
    'W': 238., 'Y': 204.,
    'U': 0.0, 'X': 0.0  # Invented
}

# Hydrophobicity
# Source: Wimley and White, Nat Struct Biol 3:842 (1996),
#         http://www.cgl.ucsf.edu/chimera/docs/UsersGuide/midas/hydrophob.html
_AA_Hww = {
    "D": -1.23, "E": -2.02, "N": -0.42, "Q": -0.58, "K": -0.99, "R": -0.81,
    "H": -0.96, "G": -0.01, "P": -0.45, "S": -0.13, "T": -0.14, "C":  0.24,
    "M":  0.23, "A": -0.17, "V": -0.07, "I":  0.31, "L":  0.56, "F":  1.13,
    "W":  1.85, "Y":  0.94,
    'U': 0.0, 'X': 0.0  # Invented
}

# Free energies of transfer between Octanol and water
# Source: Octanol and water. By Fauchere & Pliska.(1989)
_AA_Hwo = {
    "A": 0.42, "C":  1.34, "D": -1.05, "E": -0.87, "F": 2.44, "G":  0.00,
    "H": 0.18, "I":  2.46, "K": -1.35, "L":  2.32, "M": 1.68, "N": -0.82,
    "P": 0.98, "Q": -0.30, "R": -1.37, "S": -0.05, "T": 0.35, "V":  1.66,
    "W": 3.07, "Y":  1.31,
    'U': 0.0, 'X': 0.0  # Invented
}


# Hydrophaty Index
# Source: A simple method for displaying the hydropathic character of a
#         protein. Kyte and Doolittle
_AA_Hkd = {
    "F":  2.8, "I":  4.5, "W": -0.9, "L":  3.8, "V":  4.2, "M":  1.9,
    "Y":  6.3, "C":  2.5, "A":  1.8, "T": -1.3, "H": -3.2, "G": -0.4,
    "S": -0.8, "Q": -3.5, "R": -4.5, "K": -3.9, "N": -3.5, "E": -3.5,
    "P": -1.6, "D": -3.5, "X": -0.14,
    'U': 0.0, 'X': 0.0  # Invented
}


def _diff(variant, values_dict):
    return values_dict[variant.wt] - values_dict[variant.mt]


def _rdiff(variant, values_dict):
    diff = _diff(variant, values_dict)
    if values_dict[variant.wt] == 0:
        return diff / sys.float_info.epsilon
    return diff / values_dict[variant.wt]


def _kyte_doolittle(variant, window_size=9):
    start = (variant.position-1) - (window_size - 1)//2
    if start < 0:
        start = 0

    end = (variant.position-1) + (window_size + 1)//2
    if end > len(variant.sequence):
        end = len(variant.sequence)

    sum_wt = sum([_AA_Hkd[aa] for aa in variant.sequence[start:end]])
    sum_mt = sum_wt - _AA_Hkd[variant.wt] + _AA_Hkd[variant.mt]

    return float(sum_wt - sum_mt) / (end - start)


def _matrix(variant, matrix):
    if (variant.wt, variant.mt) in matrix:
        return matrix[(variant.wt, variant.mt)]
    elif (variant.mt, variant.wt) in matrix:
        return matrix[(variant.mt, variant.wt)]
    else:
        return np.nan


def _basic_feature(variants, feature):
    if feature == 'feature_pos':
        return variants['position']
    elif feature == 'feature_hkd':
        return variants.apply(_kyte_doolittle, axis=1)
    elif feature == 'feature_vol_d':
        return variants.apply(_diff, args=(_AA_Vol,), axis=1)
    elif feature == 'feature_vol_r':
        return variants.apply(_rdiff, args=(_AA_Vol,), axis=1)
    elif feature == 'feature_hww_d':
        return variants.apply(_diff, args=(_AA_Hww,), axis=1)
    elif feature == 'feature_hww_r':
        return variants.apply(_rdiff, args=(_AA_Hww,), axis=1)
    elif feature == 'feature_hwo_d':
        return variants.apply(_diff, args=(_AA_Hwo,), axis=1)
    elif feature == 'feature_hwo_r':
        return variants.apply(_rdiff, args=(_AA_Hwo,), axis=1)

    m = re.match('feature_(blosum80|blosum62|blosum50|pam60|miyata)', feature)
    if m:
        matrix = _SUBSTITUTION_MATRICES[m.group(1)]
        return variants.apply(_matrix, args=(matrix,), axis=1)


def _basic_features(variants, features):
    logging.info('Computing basic features...')
    for feature in set(features).intersection(_FEATURES['basic']):
        variants[feature] = _basic_feature(variants, feature)


#####################
# Topology features #
#####################

_TOPOLOGY_FEATURES = {}
_TOPOLOGY_FEATURES_INDEX = {}


def _load_topology_features():
    global _TOPOLOGY_FEATURES
    global _TOPOLOGY_FEATURES_INDEX
    _TOPOLOGY_FEATURES, _TOPOLOGY_FEATURES_INDEX = get_network_parameters()


def _topology_features(variants, features):
    logging.info('Computing topology features...')
    _load_topology_features()
    for feature in set(features).intersection(_FEATURES['topology']):
        variants[feature] = _topology_feature(variants, feature)


def _topology_feature(variants, feature):
    index = _TOPOLOGY_FEATURES_INDEX[feature]
    return variants.apply(_topology_variant_feature, args=(index,), axis=1)


def _topology_variant_feature(variant, index):
    if variant.protein_id in _TOPOLOGY_FEATURES:
        return _TOPOLOGY_FEATURES[variant.protein_id][index]
    return np.nan


##################
# Blast features #
##################

_BLAST_EVALUE = {
    '90': 1e-45,
    '100': 1e-75,
    'test': 1e-75,
}


def _blast_features(variants, features):
    logging.info('Computing blast features for %d variants...' % len(variants))
    for db, blast_features in _BLAST_FEATURES.items():
        db_features = set(features).intersection(blast_features)
        if len(db_features) > 0:
            for protein_id, group in variants.groupby('protein_id'):
                _blast_protein_features(variants, group.index, db_features, db)


def _blast_protein_features(variants, indices, features, db):
    # Get blast
    protein_id, sequence = variants.loc[indices[0]][['protein_id', 'sequence']]
    blast = _get_blast(protein_id, sequence, db)

    # Compute features for each position
    for position, group in variants.loc[indices].groupby('position'):
        _blast_position_features(variants, group.index, features, db, blast)


def _blast_position_features(variants, indices, features, db, blast):
    # Get position column
    position = variants.loc[indices[0]].position
    try:
        aas = blast['aas'][:, position - 1]
        scores = blast['scores'][:, position - 1]
        evalue = blast['evalue'][:, position - 1]
    except:
        aas = np.array([], dtype=np.character)
        scores = np.array([], dtype=np.float)
        evalue = np.array([], dtype=np.float)

    aas_evalue = aas[evalue < _BLAST_EVALUE[db]]
    weights_evalue = scores[evalue < _BLAST_EVALUE[db]]
    aas_human = aas[blast['human']]
    weights_human = scores[blast['human']]
    aas_not_human = aas[np.logical_not(blast['human'])]
    weights_not_human = scores[np.logical_not(blast['human'])]

    for feature in features:
        feature_parts = feature.split('_')
        subset, aas_feature = feature_parts[2:4]

        if subset == 'all':
            weights = scores if feature_parts[-1] == 'w' else None
            for i in indices:
                variants.set_value(i, feature, _aas_feature(
                    variants.loc[i], aas_feature, aas, weights))
        elif subset == 'eva':
            weights = weights_evalue if feature_parts[-1] == 'w' else None
            for i in indices:
                variants.set_value(i, feature, _aas_feature(
                    variants.loc[i], aas_feature, aas_evalue, weights))
        elif subset == 'hum':
            weights = weights_human if feature_parts[-1] == 'w' else None
            for i in indices:
                variants.set_value(i, feature, _aas_feature(
                    variants.loc[i], aas_feature, aas_human, weights))
        elif subset == 'nhu':
            weights = weights_not_human if feature_parts[-1] == 'w' else None
            for i in indices:
                variants.set_value(i, feature, _aas_feature(
                    variants.loc[i], aas_feature, aas_not_human, weights))


def _aas_feature(variant, feature, aas, weights):
    if weights is None:
        weights = np.repeat(1., len(aas))

    def n_aln():
        return weights.sum()

    def wt_aas():
        return float(((aas == variant.wt.encode())*weights).sum())

    def mt_aas():
        return float(((aas == variant.mt.encode())*weights).sum())

    def difference_to_property_mean(property_dict):
        property_sum = 0.0
        property_n = 0.0
        for k, v in property_dict.items():
            n = (aas == k.encode()).sum()
            property_n += n
            property_sum += n*v
        return property_sum/property_n - property_dict[variant.mt]

    if feature == 'nal':
        return n_aln()
    elif feature == 'naa':
        return float(((aas != b'-')*weights).sum())
    elif feature == 'nwt':
        return wt_aas()
    elif feature == 'nmt':
        return mt_aas()
    elif feature == 'rwt':
        return wt_aas() / len(aas) if len(aas) > 0 else 0.0
    elif feature == 'rmt':
        return mt_aas() / len(aas) if len(aas) > 0 else 0.0
    elif feature == 'pwm':
        return _pssm(variant, wt_aas(), mt_aas())
    elif feature == 'dh':
        return difference_to_property_mean(_AA_Hkd)


def _pssm(variant, wt_x, mt_x):
    try:
        return (math.log(mt_x/_AA_Freq[variant.mt]) -
                math.log(wt_x/_AA_Freq[variant.wt]))
    except ValueError:
        return -10.


# Composition in percent for the complete Swissprot database
# (only Organism: Human)
_AA_Freq = {
    "A": 7.015, "C": 2.299, "E": 7.094, "D": 4.737, "G": 6.577, "F": 3.655,
    "I": 4.340, "H": 2.631, "K": 5.723, "M": 2.131, "L": 9.962, "N": 3.590,
    "Q": 4.764, "P": 6.306, "S": 8.318, "R": 5.645, "T": 5.356, "W": 1.220,
    "V": 5.973, "Y": 2.666,
    'U': sys.float_info.min, 'X': sys.float_info.min  # Invented
}


###################
# Kalign features #
###################

def _kalign_features(variants, features):
    logging.info('Computing kalign features for %d variants...' %
                 len(variants))
    for db, kalign_features in _KALIGN_FEATURES.items():
        db_features = set(features).intersection(kalign_features)
        if len(db_features) > 0:
            for protein_id, group in variants.groupby('protein_id'):
                _kalign_protein_features(
                    variants, group.index, db_features, db)


def _kalign_protein_features(variants, indices, features, db):
    # Get kalign
    protein_id, sequence = variants.loc[indices[0]][['protein_id', 'sequence']]
    kalign = _get_kalign(protein_id, sequence, db)

    # Compute features for each position
    for position, group in variants.loc[indices].groupby('position'):
        _kalign_position_features(variants, group.index, features, db, kalign)


def _kalign_position_features(variants, indices, features, db, kalign):
    # Get position column
    position = variants.loc[indices[0]].position
    aas = kalign['aligned_aas'][:, position - 1]

    aas_human = aas[kalign['human']]
    weights_human = kalign['similarity'][kalign['human']]
    aas_not_human = aas[np.logical_not(kalign['human'])]
    weights_not_human = kalign['similarity'][np.logical_not(kalign['human'])]

    for feature in features:
        feature_parts = feature.split('_')
        subset, aas_feature = feature_parts[2:4]

        if subset == 'all':
            weights = kalign['similarity'] if feature_parts[-1] == 'w' \
                else None
            for i in indices:
                variants.set_value(i, feature, _aas_feature(
                    variants.loc[i], aas_feature, aas, weights))
        elif subset == 'hum':
            weights = weights_human if feature_parts[-1] == 'w' else None
            for i in indices:
                variants.set_value(i, feature, _aas_feature(
                    variants.loc[i], aas_feature, aas_human, weights))
        elif subset == 'nhu':
            weights = weights_not_human if feature_parts[-1] == 'w' else None
            for i in indices:
                variants.set_value(i, feature, _aas_feature(
                    variants.loc[i], aas_feature, aas_not_human, weights))


#############
# Get BLAST #
#############

def _get_blast_alignments(f):
    # zcat UniRef100_P13807.100.xml.gz
    #  | awk '/<Iteration>/{i++}i==2{print}'
    #  | grep 'Hit_def\|query-from\|query-to\|hseq\|qseq\|bit-score\|evalue'
    #  | sed 's/<Hit_def>/<Hit_def>!!!/g'
    #  | awk -F '[<>]' '//{print $3}'
    Alignment = collections.namedtuple('Alignment', 'title, hsps')
    Hsp = collections.namedtuple(
        'Hsp', 'query_start, query_end, sbjct, query, bits, expect')

    re_inside_tags = re.compile(b'>(.*)<')

    def get_field(s):
        return re_inside_tags.search(line).groups()[0]

    title = ''
    hsps = []
    for line in f:
        if line == b'</Iteration>\n':
            break

    for line in f:
        if line == b'<Hit>\n':
            for line in f:
                if line.startswith(b'  <Hit_def>'):
                    title = line.split(b'>')[1].split(b'<')[0].decode()
                if line == b'  <Hit_hsps>\n':
                    hsps = []
                    for line in f:
                        if line == b'    <Hsp>\n':
                            hsp = {}
                            for line in f:
                                if line.startswith(b'      <Hsp_query-from>'):
                                    hsp['query_start'] = int(get_field(line))
                                if line.startswith(b'      <Hsp_query-to>'):
                                    hsp['query_end'] = int(get_field(line))
                                if line.startswith(b'      <Hsp_hseq>'):
                                    hsp['sbjct'] = get_field(line)
                                if line.startswith(b'      <Hsp_qseq>'):
                                    hsp['query'] = get_field(line)
                                if line.startswith(b'      <Hsp_bit-score>'):
                                    hsp['bits'] = float(get_field(line))
                                if line.startswith(b'      <Hsp_evalue>'):
                                    hsp['expect'] = float(get_field(line))
                                if line == b'    </Hsp>\n':
                                    break
                            hsps.append(Hsp(**hsp))
                        elif line == b'  </Hit_hsps>\n':
                            break
                elif line == b'</Hit>\n':
                    break
            yield Alignment(title, hsps)


def _get_blast(protein_id, sequence, db):
    with io.BufferedReader(gzip.open(path_blast(protein_id, db), 'rb')) as f:
        aas = []
        scores = []
        evalue = []
        human = []

        for alignment in _get_blast_alignments(f):
            # New row
            new_aas = np.array([b'-']*len(sequence), dtype=np.character)
            new_scores = np.full(len(sequence), 0.0)
            new_evalue = np.full(len(sequence), 1.0)
            new_human = 'Homo sapiens' in alignment.title

            for hsp in alignment.hsps:
                start = hsp.query_start - 1
                end = hsp.query_end

                if (hsp.query == hsp.sbjct and
                        start == 0 and end == len(sequence)):
                    pass  # Exact match

                if (new_aas[start:end] != b'-').any():
                    # Append row
                    aas.append(new_aas)
                    scores.append(new_scores)
                    evalue.append(new_evalue)
                    human.append(new_human)

                    # New row
                    new_aas = np.array([b'-']*len(sequence),
                                       dtype=np.character)
                    new_scores = np.full(len(sequence), 0.0)
                    new_evalue = np.full(len(sequence), 1.0)
                    new_human = 'Homo sapiens' in alignment.title

                new_aas[start:end] = [
                    chr(s)
                    for q, s in zip(hsp.query, hsp.sbjct)
                    if q != ord(b'-')]
                new_scores[start:end] = hsp.bits  # Use bit score (normalized)
                new_evalue[start:end] = hsp.expect

            # Append row
            aas.append(new_aas)
            scores.append(new_scores)
            evalue.append(new_evalue)
            human.append(new_human)

    return {
        'aas': np.array(aas, dtype=np.character, order='F'),
        'scores': np.array(scores, order='F'),
        'evalue': np.array(evalue, order='F'),
        'human': np.array(human, dtype=np.bool),
    }


##############
# Get Kalign #
##############

def _get_kalign(protein_id, sequence, db):
    with io.BufferedReader(
            gzip.open(path_kalign(protein_id, db), 'rb')) as f:
        # Find protein in alignment
        self_name = uniref100(protein_id)
        aas = []
        names = []
        titles = []
        for title, seq in load_fasta_iter(f):
            titles.append(title)
            if '|' in title:
                name = title.split('|')[1].split()[0]
            else:
                name = title
            names.append(name)
            if self_name == name:
                # Protein sequence amino acid positions in alignment
                self_seq = seq
                seq_indices = np.where(seq != b'-')[0]
            aas.append(seq)

        if self_name not in names:
            raise KeyError('%s not found in %s.%s.afa.gz' % (
                self_name, self_name, db))

        similarity = np.array([
            float(((self_seq == row) & (row != b'-')).sum()) /
            len(sequence)
            for row in aas
        ])

        return {
            'aligned_aas': np.array(
                [seq[seq_indices] for seq in aas], dtype=np.character),
            'similarity': similarity,
            'names': names,
            'human': np.array([
                'Homo sapiens' in title for title in titles],
                dtype=np.bool)
        }


#########################
# Features distribution #
#########################

def features_distribution(variants, features=None):
    if features is None:
        features = [c for c in variants.columns if 'feature_' in c]

    sns.set_style("white")
    figures = []

    for feature in features:
        fig = plt.figure()

        plt.hist(
            variants[
                np.logical_not(variants.disease)][feature].dropna().values,
            label='Neutral', bins=20, alpha=0.5, color='g', normed=True,
            range=((variants[feature].min(), variants[feature].max())))
        plt.hist(variants[variants.disease][feature].dropna().values,
                 label='Disease', bins=20, alpha=0.5, color='r', normed=True,
                 range=((variants[feature].min(), variants[feature].max())))

        plt.legend(loc='upper right')
        plt.title(feature_name(feature), y=1.02)
        plt.locator_params(axis='y', nbins=6)

        figures.append((feature_name(feature), fig))

    return figures




import re

import numpy as np
import pandas as pd

from .amino_acids import AMINO_ACIDS, AA_DISTANCE, aa1


def variants():
    return pd.DataFrame(columns=[
        'protein_id', 'sequence', 'position', 'wt', 'mt', 'disease'])


def variant_str(variant):
    return '%s %d.%s>%s' % (
        variant.protein_id, variant.position, variant.wt, variant.mt)


def load_fasta_str(f):
    return {
        descr: b''.join(sequence).decode()
        for descr, sequence in load_fasta_iter(f)
    }


def load_fasta_iter(f):
    descr = ''
    sequence = b''

    for line in f:
        line = line.strip()
        if line == b'':
            continue
        if line[0] == ord(b'>'):
            if descr:
                yield (descr, np.array([chr(c) for c in sequence],
                       dtype=np.character))
                descr = ''
                sequence = ''
            descr = line[1:].decode()
            sequence = b''
        else:
            sequence += line

    if descr:
        yield (descr, np.array([chr(c) for c in sequence], dtype=np.character))


def all_variants(protein_id, sequence, position=None):
    """Position is 0-based"""
    positions = [(position, sequence[position])] \
        if position else enumerate(sequence)

    return pd.DataFrame([
            [protein_id, sequence, i+1, wt, mt]
            for i, wt in positions
            for mt in AMINO_ACIDS
            if mt != wt
        ], columns=['protein_id', 'sequence', 'position', 'wt', 'mt'])


def parse_variant(string):
    p = re.compile(r'p.([A-Z][a-z]{2})(\d+)([A-Z][a-z]{2})')
    m = p.match(string)

    assert m and m.group() == string, '%s cannot be parsed as variant' % string

    return {
        'position': int(m.group(2)),
        'wt': aa1(m.group(1)),
        'mt': aa1(m.group(3)),
    }


def position_score(wt, position_scores):
    scores_at_dist_1 = [
        sc
        for mt, sc in position_scores.items()
        if sc and AA_DISTANCE[wt][mt] == 1
    ]
    if not scores_at_dist_1:
        return np.median(list(position_scores.values()))

    return np.median(scores_at_dist_1) if scores_at_dist_1 else None



# -*- coding: utf-8 -*-

import csv
import io
import logging
import os
import re
import subprocess

from .utils import load_fasta_str

# Paths

_PROJECT_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), os.pardir)
_DATA_DIR = os.getenv('PMUT_DATA_DIR', os.path.join(_PROJECT_DIR, 'data'))
_VENDOR_DIR = os.path.join(_PROJECT_DIR, 'vendor')

_HUMAN_UNIREF100 = os.path.join(
    _VENDOR_DIR, 'Uniref100_human', 'UniRef100_human')
_BLASTDBCMD = os.path.join(
    _VENDOR_DIR, 'blastdbcmd')

##############################
# Blast, Fa and Kalign paths #
##############################


def _return_fname_if_exists(fname):
    if os.path.isfile(fname):
        return fname
    else:
        raise Exception('File not found: %s' % fname)


def path_fasta(protein_id, db):
    """Get path of fasta file, raise Exception if not found"""
    return _return_fname_if_exists(os.path.join(
        _DATA_DIR, 'fasta', uniref100(protein_id) + '.fasta'))


def _get_file(name, protein_id, db, extension):
    return _return_fname_if_exists(os.path.join(
        _DATA_DIR, '%s_%s' % (name, db),
        '%s.%s.%s' % (uniref100(protein_id), db, extension)))


def path_blast(protein_id, db):
    """Get path of blast file, raise Exception if not found"""
    return _get_file('blast', protein_id, db, 'xml.gz')


def path_fa(protein_id, db):
    """Get path of fa file, raise Exception if not found"""
    return _get_file('fa', protein_id, db, 'fa.gz')


def path_kalign(protein_id, db):
    """Get path of kalign file, raise Exception if not found"""
    return _get_file('kalign', protein_id, db, 'afa.gz')


##################################
# UniProt -> UniRef100, UniRef50 #
##################################

_UNIREF100_2_UNIPROT_FILE = os.path.join(
    _VENDOR_DIR, 'human_uniref100_2_uniprot.txt')

_UNIREF50_2_UNIPROT_FILE = os.path.join(
    _VENDOR_DIR, 'human_uniref50_2_uniprot.txt')  # NOTE: only human

_UNIPROT_2_UNIREF100 = {}
_UNIPROT_2_UNIREF50 = {}


def _load_uniref_to_uniprot(fname):
    with open(fname) as f:
        return {
            prot.strip(): line.split('\t')[0]
            for line in f.read().splitlines()
            for prot in line.split('\t')[1].split('; ')
        }


def uniref100(uniprot):
    """Return uniref100 cluster for uniprot if found,
    return uniprot otherwise"""
    global _UNIPROT_2_UNIREF100
    if not _UNIPROT_2_UNIREF100:
        _UNIPROT_2_UNIREF100 = _load_uniref_to_uniprot(
            _UNIREF100_2_UNIPROT_FILE)
    return _UNIPROT_2_UNIREF100.get(uniprot, uniprot)


def uniref50(uniprot):
    """Return uniref100 cluster for uniprot if found,
    return uniprot otherwise"""
    global _UNIPROT_2_UNIREF50
    if not _UNIPROT_2_UNIREF50:
        _UNIPROT_2_UNIREF50 = _load_uniref_to_uniprot(_UNIREF50_2_UNIPROT_FILE)
    return _UNIPROT_2_UNIREF50.get(uniprot, uniprot)

######################
# Network parameters #
######################

_NETWORK_PARAM_FILE = os.path.join(
    _VENDOR_DIR, 'topological_characteristics_binary_interaction_data.tsv')


def get_network_parameters():
    with open(_NETWORK_PARAM_FILE) as f:
        reader = csv.reader(f, delimiter='\t')
        header = next(reader)
        return {
            row[0]: [float(x) for x in row[1:]]
            for row in csv.reader(f, delimiter='\t')
        }, {
            'feature_top_' + h: i for i, h in enumerate(header[1:])
        }


######################
# Blast DB sequences #
######################

def get_protein_sequences(protein_ids):
    uniref100_ids = [uniref100(protein_id) for protein_id in protein_ids]
    p = subprocess.Popen([
        _BLASTDBCMD, '-db', _HUMAN_UNIREF100, '-dbtype', 'prot',
        '-entry_batch', '/proc/self/fd/0'],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE)
    out, _ = p.communicate(('\n'.join(uniref100_ids)).encode())
    fasta = {
        title.split('|')[1].split()[0]: sequence
        for title, sequence in load_fasta_str(io.BytesIO(out)).items()
    }
    missing = set(uniref100_ids).difference(fasta.keys())
    if missing:
        logging.warn('Sequence not found for %d ids: %s', len(missing),
                     str(list(missing)))
    return {
        protein_id: fasta[uniref100(protein_id)]
        for protein_id in protein_ids
        if uniref100(protein_id) in fasta
    }




import logging

import pandas as pd

from .utils import parse_variant


def read_humsavar_txt(f):
    """Read a humsavar.txt file and return a variants DataFrame"""
    SWISSPROT_POS, AA_CHANGE_POS, TYPE_POS = (1, 3, 4)

    variants = set()
    header = True

    for line in f:
        line = line.strip()

        # Ignore empty lines
        if line == '':
            continue

        # Log header info
        if header:
            if '________' in line:
                header = False
            if ('Release:' in line or 'Disease variants:' in line or
                    'Polymorphisms:' in line or 'Unclassified variants:' in
                    line or 'Total:' in line):
                logging.info(line)
        else:
            # Break when the end comment is reached
            if '--------------' in line:
                break

            row = line.split()

            try:
                uniprot = row[SWISSPROT_POS]
                variant = parse_variant(row[AA_CHANGE_POS])
                typ = row[TYPE_POS]
                assert typ in ['Disease', 'Polymorphism', 'Unclassified']

                if typ == 'Disease' or typ == 'Polymorphism':
                    variants.add((
                        uniprot, variant['position'],
                        variant['wt'], variant['mt'], typ == 'Disease'))
            except:
                # Print line if error when parsing
                logging.warning('Can not parse line: %s', line)

    return pd.DataFrame(
        list(variants),
        columns=('protein_id', 'position', 'wt', 'mt', 'disease'))



import logging
import sys

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s %(levelname)-8s %(message)s',
    datefmt='%m-%d %H:%M:%S',
    stream=sys.stderr)

# Load all modules

from .amino_acids import *
from .data import *
from .features import *
from .io import *
from .pmut import *
from .prediction import *
from .utils import *



# -*- coding: utf-8 -*-

import logging

from .data import get_protein_sequences
from .utils import variant_str


def get_sequences(variants):
    protein_ids = set(variants.protein_id)
    sequences = get_protein_sequences(protein_ids)
    variants['sequence'] = [
        sequences[row.protein_id] if row.protein_id in sequences else None
        for i, row in variants.iterrows()
    ]

    mask = []
    for i, variant in variants.iterrows():
        if variant.position > len(variant.sequence):
            logging.warning(
                'Variant %s out of sequence. Sequence length: %d',
                variant_str(variant), variant.position)
            mask.append(False)
        elif variant.sequence[variant.position - 1] != variant.wt:
            logging.warning(
                'Variant %s does not match sequence. WT should be %s',
                variant_str(variant), variant.sequence[variant.position - 1])
            mask.append(False)
        else:
            mask.append(True)

    return variants[mask]



# -*- coding: utf-8 -*-

from distutils.core import setup

setup(
    name = 'pymut',
    packages = ['pymut'],
    version = '0.1.1',
    description = 'Machine learning methods for the prediction of pathology in protein mutations, as in the PMut predictor (http://mmb.pcb.ub.es/pmut2017/)',
    author = 'Víctor López Ferrando',
    author_email = 'victorlopez90@gmail.com',
    license = 'MIT',
    url = 'https://github.com/vlopezferrando/pymut',
    download_url = 'https://github.com/vlopezferrando/pymut/tarball/0.1.1',
    keywords = ['protein', 'mutations', 'machine learning'],
    install_requires=[
        'numpy',
        'scipy',
        'pandas',
        'matplotlib',
        'seaborn',
        'scikit-learn',
        'imbalanced-learn',
    ],
)
